/*
 * Copyright 2008-2023 Aerospike, Inc.
 *
 * Portions may be licensed to Aerospike, Inc. under one or more contributor
 * license agreements.
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 */
 
#pragma once

#include <aerospike/as_std.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>

bool
as_tls_match_name(X509 *x509, const char *name, bool allow_wildcard);

#if !defined(_MSC_VER)
bool
as_tls_match_san(X509 *x509, const char *name);
#endif

